#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2016 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Report card view components
"""

import datetime
import pytz

from reportlab.lib import units
from reportlab.lib import pagesizes
from z3c.form import field
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.interface import Interface
from zope.schema import Choice
from zope.security.proxy import removeSecurityProxy

from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.membership import Membership
from schooltool.ark import ARK_DATE_FORMAT
from schooltool.ark.browser.assessment import AssessmentReportBase
from schooltool.ark.browser.report import filter_students_requirements
from schooltool.ark.browser.report import getScoreSystem
from schooltool.common import SchoolToolMessage as STMessage
from schooltool.common import DateRange
from schooltool.common import getResourceURL
from schooltool.course.interfaces import ICourseContainer
from schooltool.course.interfaces import ISection
from schooltool.gradebook.interfaces import IActivities
from schooltool.gradebook.interfaces import IGradebook
from schooltool.level.interfaces import ILevelContainer
from schooltool.lyceum.journal.journal import StudentAMRequirement
from schooltool.lyceum.journal.journal import StudentPMRequirement
from schooltool.lyceum.journal.browser.journal import makeSchoolAttendanceMeeting
from schooltool.lyceum.journal.interfaces import IAttendanceOptions
from schooltool.lyceum.journal.interfaces import IEvaluateRequirement
from schooltool.lyceum.journal.interfaces import IJournalScoreSystemPreferences
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin import flourish
from schooltool.stream.interfaces import IStreamContainer
from schooltool.stream.stream import StreamMembers
from schooltool.term.interfaces import IDateManager
from schooltool.term.interfaces import ITerm

from schooltool.rising import RisingMessage as _


class ITermReportCardForm(Interface):

    term = Choice(
        title=_('Term'),
        source='schooltool.ark.year_terms')

    level = Choice(
        title=STMessage('Level'),
        vocabulary='schooltool.level.level.levelvocabulary')

    stream = Choice(
        title=STMessage('Stream'),
        source='schooltool.ark.streams')


class TermReportCardRequestView(RequestRemoteReportDialog):

    report_builder = 'term_report_card.pdf'

    fields = field.Fields(ITermReportCardForm)

    @Lazy
    def schoolyear(self):
        return self.context

    def resetForm(self):
        RequestRemoteReportDialog.resetForm(self)
        self.form_params['schoolyear'] = self.schoolyear

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term.__name__
        level = self.form_params.get('level')
        task.request_params['level'] = level.__name__
        stream = self.form_params.get('stream')
        if stream is not u'':
            task.request_params['stream'] = stream.__name__

    def updateWidgets(self):
        super(TermReportCardRequestView, self).updateWidgets()
        self.widgets['term'].prompt = True
        self.widgets['level'].prompt = True


class TermReportCardPlainPageTemplate(flourish.report.PlainPageTemplate):

    @Lazy
    def header_school_logo(self):
        default = super(
            TermReportCardPlainPageTemplate, self).header_school_logo
        doc_w, doc_h = self.manager.page_size
        if default is None:
            return None
        default['x'] = self.manager.margin.left
        default['y'] = doc_h - (self.manager.margin.top + default['height'])
        return default

    @Lazy
    def frame(self):
        doc_w, doc_h = self.manager.page_size
        width = (doc_w - self.manager.margin.left - self.manager.margin.right)
        height = (doc_h - self.manager.margin.top - self.manager.margin.bottom)
        x = self.manager.margin.left
        y = self.manager.margin.bottom
        return {
            'height': height,
            'margin': flourish.report.Box(0),
            'width': width,
            'x': x,
            'y': y,
            }


class TermReportCardPDFView(flourish.report.PlainPDFPage,
                            AssessmentReportBase):

    page_size = pagesizes.landscape(pagesizes.A4)
    margin = flourish.report.Box(0.5 * units.cm, 0.7 * units.cm)

    def days_absent(self, student, term):
        result = []
        graded = []
        if term.first > self.current_term.first:
            return
        meetings = self.meetings[term]
        for meeting_data in meetings:
            requirements = meeting_data['requirements']
            for requirement in requirements:
                score = IEvaluateRequirement(requirement).getEvaluation(
                    student, requirement, default=UNSCORED)
                if score not in (UNSCORED, None):
                    graded.append(1)
                    if self.isAbsent(score):
                        result.append(1)
        if graded:
            return {
                'value': sum(result) / len(self.student_requirement_factories),
                'count': sum(graded) / len(self.student_requirement_factories),
            }

    def days_late(self, student, term):
        result = []
        graded = []
        if term.first > self.current_term.first:
            return
        meetings = self.meetings[term]
        for meeting_data in meetings:
            requirements = meeting_data['requirements']
            for requirement in requirements:
                score = IEvaluateRequirement(requirement).getEvaluation(
                    student, requirement, default=UNSCORED)
                if score not in (UNSCORED, None):
                    graded.append(1)
                    if self.isLate(score):
                        result.append(1)
        if graded:
            return {
                'value': sum(result) / len(self.student_requirement_factories),
                'count': sum(graded) / len(self.student_requirement_factories),
            }

    def days_suspended(self, student, term):
        result = []
        graded = []
        if term.first > self.current_term.first:
            return
        meetings = self.meetings[term]
        for meeting_data in meetings:
            requirements = meeting_data['requirements']
            for requirement in requirements:
                score = IEvaluateRequirement(requirement).getEvaluation(
                    student, requirement, default=UNSCORED)
                if score not in (UNSCORED, None):
                    graded.append(1)
                    if self.isSuspended(score):
                        result.append(1)
        if graded:
            return {
                'value': sum(result) / len(self.student_requirement_factories),
                'count': sum(graded) / len(self.student_requirement_factories),
            }

    def isAbsent(self, score):
        if score in (UNSCORED, None) or not score:
            return False
        if score.requirement in self.scoresystem_cache:
            ss = self.scoresystem_cache[score.requirement]
        else:
            ss = getScoreSystem(score)
            self.scoresystem_cache[score.requirement] = ss
        return ss.isAbsent(score)

    def isLate(self, score):
        if score in (UNSCORED, None) or not score:
            return False
        return score.value in self.students_timestamp_codes

    def isSuspended(self, score):
        if score in (UNSCORED, None) or not score:
            return False
        if score.requirement in self.scoresystem_cache:
            ss = self.scoresystem_cache[score.requirement]
        else:
            ss = getScoreSystem(score)
            self.scoresystem_cache[score.requirement] = ss
        return 'suspended' in ss.scoresDict.get(score.value, '').lower()

    def face(self, data):
        percent = 100 - ((data['value'] * 100.0) / data['count'])
        if percent >= 97:
            return 'smilingface'
        elif percent >= 93:
            return 'neutralface'
        else:
            return 'sadface'

    @Lazy
    def student_requirement_factories(self):
        return filter_students_requirements([
            StudentAMRequirement, StudentPMRequirement
        ])

    @Lazy
    def journal_preferences(self):
        app = ISchoolToolApplication(None)
        return IJournalScoreSystemPreferences(app)

    @Lazy
    def students_timestamp_codes(self):
        options = IAttendanceOptions(ISchoolToolApplication(None))
        return options.students_timestamp_codes or ()

    @Lazy
    def student_scoresystem(self):
        return self.journal_preferences.student_scoresystem

    @Lazy
    def meetings(self):
        result = {}
        ignore_future = True
        today = getUtility(IDateManager).today        
        ss = self.student_scoresystem
        for term in self.terms:
            term_meetings = []
            for date in term:
                if ignore_future and date > today:
                    continue
                if term.isSchoolday(date):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day)
                    )
                    meeting = makeSchoolAttendanceMeeting(dt)
                    requirements = [
                        factory(meeting, ss)
                        for factory in self.student_requirement_factories
                    ]
                    term_meetings.append({
                        'meeting': meeting,
                        'requirements': requirements,
                    })
            result[term] = term_meetings
        return result

    @Lazy
    def assessment_data(self):
        result = {}
        for student in self.students:
            result[student.username] = student_grades = {}
            sections = self.get_sections(student)
            for section in sections:
                activities = IActivities(section)
                term = ITerm(section)
                for course in section.courses:
                    averages = {}
                    for sheet, weight in self.term_sheets(term, course).items():
                        ws = activities.get(sheet)
                        if ws is not None:
                            gb = IGradebook(ws)
                            total, average = gb.getWorksheetTotalAverage(
                                ws, student)
                            if average not in [None, UNSCORED]:
                                averages[sheet] = float(average)
                    if averages:
                        key = course.__name__, term.__name__
                        student_grades[key] = {
                            'averages': averages,
                            'overall_average': self.overall_average(averages),
                        }
                         
        return result

    def get_sections(self, student):
        rel = Membership.bind(member=student).all().relationships
        result = set()
        for link_info in rel:
            if not ISection.providedBy(link_info.target):
                continue
            section = removeSecurityProxy(link_info.target)
            section_term = ITerm(section)
            section_schoolyear = ISchoolYear(section_term)
            if section_schoolyear != self.schoolyear:
                continue
            if section_term.first > self.current_term.first:
                continue
            result.add(section)
        return result

    @Lazy
    def attendance_colWidths(self):
        total_width = 12.5
        result = [
            8,
        ]
        available = total_width - sum(result)
        n = len(self.terms) * 2
        result.extend([available/float(n)] * n)
        return ','.join(['%.2f%%' % ((float(w)*100)/total_width)
                         for w in result])

    @Lazy
    def subject_colWidths(self):
        total_width = 14
        result = [
            9,
            1.2,
        ]
        available = total_width - sum(result)
        n = len(self.terms) + 1
        result.extend([available/float(n)] * n)
        return ','.join(['%.2f%%' % ((float(w)*100)/total_width)
                         for w in result])

    def averages(self, student, term, course):
        student_grades = self.assessment_data.get(student.username, {})
        key = course.__name__, term.__name__
        return student_grades.get(key, {})

    def overall_average(self, averages):
        # XXX: use weights if needed
        if averages:
            return self.div(sum(averages.values()), len(averages))

    def grade(self, student, term, course):
        if term.first > self.current_term.first:
            return
        student_grades = self.assessment_data.get(student.username, {})
        key = course.__name__, term.__name__
        course_term_grades = student_grades.get(key, {})
        overall_average = course_term_grades.get('overall_average')
        if overall_average is not None:
            return '%0.f%%' % overall_average

    def _rank(self, student, term, course):
        ranks = []
        key = course.__name__, term.__name__
        for s in self.students:
            if s == student:
                continue
            student_grades = self.assessment_data.get(s.username, {})
            course_term_grades = student_grades.get(key, {})
            overall_average = course_term_grades.get('overall_average')
            if overall_average is not None:
                ranks.append(overall_average)
        student_grades = self.assessment_data.get(student.username, {})
        course_term_grades = student_grades.get(key, {})
        overall_average = course_term_grades.get('overall_average')
        if overall_average is not None:
            ranks.append(overall_average)
            ranks = sorted(ranks, reverse=True)
            return (ranks.index(overall_average) + 1, len(ranks))

    def rank(self, student, term, course):
        if term.first > self.current_term.first:
            return
        result = self._rank(student, term, course)
        if result is not None:
            return '%d/%d' % result

    def div(self, count, total):
        if not total:
            return 0.0
        return count/float(total)

    def _overall_grade(self, student, course):
        term_averages = {}
        terms = [term for term in self.terms
                 if term.first <= self.current_term.first]
        for term in terms:
            key = course.__name__, term.__name__
            student_grades = self.assessment_data.get(student.username, {})
            course_term_grades = student_grades.get(key, {})
            overall_average = course_term_grades.get('overall_average')
            if overall_average is not None:
                term_averages[term] = overall_average
        return term_averages

    def overall_grade(self, student, course):
        term_averages = self._overall_grade(student, course)
        if term_averages:
            return '%0.f%%' % self.overall_average(term_averages)

    def section(self, student, course):
        result = set()
        for section in course.sections:
            if student in section.members.all():
                result.add(section.title)
        return ', '.join(result)

    def teacher(self, student, course):
        result = set()
        for section in course.sections:
            if student in section.members.all():
                instructor =', '.join([instructor.title
                                       for instructor in section.instructors])
                if instructor:
                    result.add(instructor)
        return ', '.join(sorted(result))

    def _overall_rank(self, student, course):
        overall_grades = []
        for s in self.students:
            if s == student:
                continue
            term_averages = self._overall_grade(s, course)
            if term_averages:
                overall_grades.append(self.overall_average(term_averages))
        term_averages = self._overall_grade(student, course)
        if term_averages:
            student_average = self.overall_average(term_averages)
            overall_grades.append(student_average)
            overall_grades = sorted(overall_grades, reverse=True)
            return (overall_grades.index(student_average) + 1,
                    len(overall_grades))

    def overall_rank(self, student, course):
        result = self._overall_rank(student, course)
        if result is not None:
            return '%d/%d' % result

    def courses(self):
        return list(ICourseContainer(self.schoolyear).values())

    @Lazy
    def terms(self):
        return sorted(self.schoolyear.values(), key=lambda t: t.first)

    @Lazy
    def current_term(self):
        return self.schoolyear.get(self.request.get('term'))

    @Lazy
    def current_term_title(self):
        return '%s, %s' % (self.current_term.title,
                           ISchoolYear(self.current_term).title)

    @Lazy
    def rising_logo(self):
        height = 3.5 * units.cm
        doc_w, doc_h = self.page_size
        illustration_width = 5 * units.cm
        return {
            'url': getResourceURL('schooltool.rising.custom',
                                  'logo_bw.png',
                                  self.request),
            'illustration_width': illustration_width,
            'x': 0,
            'y': 0,
            'height': height,
        }

    def attendance_groups(self, person, schoolyear):
        result = set()
        key = person.__name__, schoolyear.__name__
        if key in self._person_attendance_groups:
            return self._person_attendance_groups[key]
        for link_info in StreamMembers.bind(member=person).all().relationships:
            stream = link_info.target
            if (ISchoolYear(stream) == schoolyear and 
                stream.attendance_group):
                result.add(stream)
        self._person_attendance_groups[key] = result
        return result

    @Lazy
    def schoolyear(self):
        return self.context

    @Lazy
    def levels(self):
        app = ISchoolToolApplication(None)
        return ILevelContainer(app)

    @Lazy
    def streams(self):
        return IStreamContainer(self.schoolyear)

    @Lazy
    def level(self):
        return self.levels.get(self.request.get('level'))

    @Lazy
    def stream(self):
        return self.streams.get(self.request.get('stream'))

    @Lazy
    def students(self):
        date_range = DateRange(self.schoolyear.first, self.schoolyear.last)
        return self.level_students(self.level, date_range)

    def sortKey(self, student):
        streams = sorted(self.attendance_groups(student, self.schoolyear),
                         key=lambda s: s.title)
        stream = ', '.join([stream.title for stream in streams])
        return stream, student.last_name, student.first_name

    @property
    def base_filename(self):
        return 'term_report_card_%s' % self.schoolyear.__name__

    @property
    def message_title(self):
        return _('${term} report card',
                 mapping={'term': self.term.title})

    @Lazy
    def app_preferences(self):
        return IApplicationPreferences(ISchoolToolApplication(None))

    def update(self):
        super(TermReportCardPDFView, self).update()
        self._person_attendance_groups = {}
        self.scoresystem_cache = {}
        self.tzinfo = pytz.timezone(self.app_preferences.timezone)
        self.requirement_factories = filter_students_requirements(
            [StudentAMRequirement, StudentPMRequirement])

    def line_start(self, repeat):
        return '0,%d' % (repeat.index * 3)

    def line_stop(self, repeat):
        return '-1,%d' % (repeat.index * 3)


class TermReportCardPDFStory(flourish.report.PDFStory):

    template = flourish.templates.Inline('''
    <tal:loop repeat="student view/students">
      <tal:block replace="structure student/schooltool:content/term_report_card" />
    </tal:loop>
    ''')

    def students(self):
        result = sorted(self.view.students, key=self.view.sortKey)
        if self.view.stream is not None:
            result = [student for student in result
                      if student in self.view.stream.members]
        return result


class TermReportCardViewletManager(flourish.viewlet.ViewletManager):

    pass


class TermReportCardViewletBase(flourish.viewlet.Viewlet):

    @Lazy
    def pdf_view(self):
        return self.view.view


class TermReportCardStudentInfo(TermReportCardViewletBase):

    template = flourish.templates.XMLFile(
        'rml/term_report_card_student_info.pt')

    def stream(self):
        streams = sorted(self.pdf_view.attendance_groups(
            self.context, self.pdf_view.schoolyear),
                         key=lambda s: s.title)
        return ', '.join([stream.title for stream in streams])

    @Lazy
    def birth_date(self):
        if self.context.birth_date is not None:
            return self.context.birth_date.strftime(ARK_DATE_FORMAT)


class TermReportCardStudentGrades(TermReportCardViewletBase):

    template = flourish.templates.XMLFile(
        'rml/term_report_card_student_grades.pt')


class TermReportCardPageBreak(TermReportCardViewletBase):

    template = flourish.templates.XMLFile(
        'rml/term_report_card_page_break.pt')


class TermReportCardStylesPart(flourish.report.PDFPart):

    pass
