#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2015 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Export view components
"""

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.demographics import BoolFieldDescription
from schooltool.basicperson.demographics import IntFieldDescription
from schooltool.basicperson.demographics import DateFieldDescription
from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.contact.contact import Contact
from schooltool.contact.interfaces import IContact
from schooltool.contact.interfaces import IContactContainer
from schooltool.contact.interfaces import IContactDemographics
from schooltool.contact.interfaces import IContactDemographicsFields
from schooltool.export.export import Date
from schooltool.export.export import RemoteMegaExporter
from schooltool.export.export import Text
from schooltool.export.importer import ContactPersonImporter
from schooltool.export.importer import FlourishRemoteMegaImporter
from schooltool.export.importer import ImportTask
from schooltool.export.importer import RemoteMegaImporter
from schooltool.export.importer import ERROR_UNWANTED_CONTACT_DATA

class RisingExporter(RemoteMegaExporter):

    def format_contact_persons(self):
        def contact_getter(attribute):
            def getter(contact):
                person = IBasicPerson(contact.__parent__, None)
                if person is None:
                    return getattr(contact, attribute)
                if attribute == '__name__':
                    return person.username
                return ''
            return getter

        def demographics_getter(attribute):
            def getter(contact):
                demographics = IContactDemographics(contact)
                return demographics[attribute]
            return getter

        fields = [('ID', Text, contact_getter('__name__')),
                  ('First Name', Text, contact_getter('first_name')),
                  ('Last Name', Text, contact_getter('last_name')),
                  ]
        app = ISchoolToolApplication(None)
        demographics_fields = IContactDemographicsFields(app)
        for field in demographics_fields.values():
            title = field.title
            format = Text
            if isinstance(field, DateFieldDescription):
                format = Date
            getter = demographics_getter(field.name)
            fields.append((title, format, getter))


        items = []
        for person in self.context['persons'].values():
            items.append(IContact(person))
        for contact in IContactContainer(self.context).values():
            items.append(contact)

        return self.format_table(fields, items, importer='export_contacts',
                                 major_progress=(0,2))


class RisingContactPersonImporter(ContactPersonImporter):

    def applyData(self, contact, data):
        if data['__name__'] not in ISchoolToolApplication(None)['persons']:
            contact.first_name = data['first_name']
            contact.last_name = data['last_name']

    def establishContact(self, data, sh, row):
        app = ISchoolToolApplication(None)
        persons = app['persons']
        contacts = IContactContainer(app)
        name = data['__name__']
        if name in persons:
            person = persons[name]
            contact = IContact(person)
            self.applyData(contact, data)
        elif name in contacts:
            contact = contacts[name]
            self.applyData(contact, data)
        else:
            contact = Contact()
            self.applyData(contact, data)
            contacts[name] = contact

        demographics = IContactDemographics(contact)
        for n, field in enumerate(self.fields):
            if field.required:
                if isinstance(field, DateFieldDescription):
                    value = self.getDateFromCell(sh, row, n + 2)
                elif isinstance(field, BoolFieldDescription):
                    value = self.getRequiredBoolFromCell(sh, row, n + 2)
                elif isinstance(field, IntFieldDescription):
                    value = self.getRequiredIntFromCell(sh, row, n + 2)
                else:
                    value = self.getRequiredTextFromCell(sh, row, n + 2)
            else:
                if isinstance(field, DateFieldDescription):
                    value = self.getDateFromCell(sh, row, n + 2,
                                                 default=None)
                elif isinstance(field, BoolFieldDescription):
                    value = self.getBoolFromCell(sh, row, n + 2)
                elif isinstance(field, IntFieldDescription):
                    value = self.getIntFromCell(sh, row, n + 2)
                else:
                    value = self.getTextFromCell(sh, row, n + 2)
                if value == '':
                    value = None
            demographics[field.name] = value

    def process(self):
        sh = self.sheet
        persons = ISchoolToolApplication(None)['persons']
        nrows = sh.nrows
        fields = IContactDemographicsFields(ISchoolToolApplication(None))
        self.fields = list(fields.values())
        for row in range(1, nrows):
            if self.isEmptyRow(sh, row):
                continue

            num_errors = len(self.errors)
            data = {}

            data['__name__'] = self.getRequiredIdFromCell(sh, row, 0)
            if data['__name__'] is not None:
                self.validateUnicode(data['__name__'], row, 0)
            if num_errors == len(self.errors):
                if data['__name__'] not in persons:
                    data['first_name'] = self.getRequiredTextFromCell(sh, row, 1)
                    data['last_name'] = self.getRequiredTextFromCell(sh, row, 2)
                else:
                    for index in range(2):
                        value, found = self.getCellAndFound(sh, row, index + 1)
                        if value:
                            self.error(row, index + 1,
                                       ERROR_UNWANTED_CONTACT_DATA)

            if num_errors == len(self.errors):
                self.establishContact(data, sh, row)
            self.progress(row, nrows)


class RisingMegaImporter(RemoteMegaImporter):

    @property
    def importers(self):
        default = super(RisingMegaImporter, self).importers
        selected = default[:8] + default[9:]
        selected.insert(8, RisingContactPersonImporter)
        return selected


class RisingImporter(FlourishRemoteMegaImporter):

    def getTask(self):
        app = ISchoolToolApplication(None)
        return ImportTask(RisingMegaImporter, app)
