#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2015 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Journey integration components
"""

from datetime import datetime
from datetime import timedelta
from urllib import urlencode
import urllib2

from persistent import Persistent
from zope.annotation.interfaces import IAnnotations
from zope.component import adapter
from zope.component import getUtility
from zope.interface import implementer
from zope.intid.interfaces import IIntIds

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.calendar.utils import parse_date
from schooltool.lyceum.journal.interfaces import IEvaluateRequirement
from schooltool.lyceum.journal.interfaces import IJournalScoreSystemPreferences
from schooltool.lyceum.journal.browser.journal import makeSchoolAttendanceMeeting
from schooltool.lyceum.journal.journal import StudentAMRequirement
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.skin.flourish.tal import JSONDecoder
from schooltool.stream.interfaces import IStream
from schooltool.task.tasks import RemoteTask

from schooltool.rising.interfaces import IJourneyAccessUtility
from schooltool.rising.interfaces import IAttendanceGroup
from schooltool.rising.interfaces import IJourneyPreferences


ATTENDANCE_GROUP_KEY = 'schooltool.stream.attendance_group'
JOURNEY_PREFERENCES_KEY = 'schooltool.rising.journey_preferences'


class HeadRequest(urllib2.Request):

    def get_method(self):
        return 'HEAD'


class AttendanceGroup(Persistent):

    attendance_group = None


@adapter(IStream)
@implementer(IAttendanceGroup)
def getAttendanceGroup(stream):
    annotations = IAnnotations(stream)
    if ATTENDANCE_GROUP_KEY not in annotations:
        annotations[ATTENDANCE_GROUP_KEY] = AttendanceGroup()
        annotations[ATTENDANCE_GROUP_KEY].__parent__ = stream
        annotations[ATTENDANCE_GROUP_KEY].__name__ = 'attendance_group'
    return annotations[ATTENDANCE_GROUP_KEY]


class JourneyPreferences(Persistent):

    days = 5
    last_sync = None


@adapter(ISchoolToolApplication)
@implementer(IJourneyPreferences)
def getJourneyPreferences(app):
    annotations = IAnnotations(app)
    if JOURNEY_PREFERENCES_KEY not in annotations:
        annotations[JOURNEY_PREFERENCES_KEY] = JourneyPreferences()
        annotations[JOURNEY_PREFERENCES_KEY].__parent__ = app
        annotations[JOURNEY_PREFERENCES_KEY].__name__ = 'journey_preferences'
    return annotations[JOURNEY_PREFERENCES_KEY]


class JourneyCallTask(RemoteTask):

    routing_key = 'zodb.report' # XXX
    context_intid = None
    request_factory = urllib2.Request

    def __init__(self, context):
        RemoteTask.__init__(self)
        self.context = context

    def execute(self, celery_task, *args, **kwargs):
        utility = getUtility(IJourneyAccessUtility)
        request = self.request_factory(self.call_url(utility))
        request.add_header('Authorization', utility.authorization_header)
        data = urlencode(self.api_call_args())
        try:
            # XXX: for GET requests, data should be in the querystring
            response = urllib2.urlopen(request, data)
        except (urllib2.HTTPError,) as exc:
            self.handle_http_error(exc)
        else:
            decoder = JSONDecoder()
            result = decoder.decode(response.read())

    def fail(self, celery_task, result, traceback):
        super(JourneyCallTask, self).fail(celery_task, result, traceback)

    @property
    def context(self):
        int_ids = getUtility(IIntIds)
        obj = int_ids.queryObject(self.context_intid)
        return obj

    @context.setter
    def context(self, value):
        int_ids = getUtility(IIntIds)
        intid = int_ids.getId(value)
        self.context_intid = intid

    def api_call_args(self):
        raise NotImplemented('subclass')

    def call_url(self, access_utility):
        raise NotImplemented('subclass')

    def handle_http_error(self, exc):
        raise exc


class Register(object):

    def __init__(self, register_info):
        self.id = register_info['id']
        self.completed_at = register_info['completed_at']
        self.edited_at = register_info['edited_at']
        self.user_id = register_info['user_id']
        self.date = register_info['date']


class StudentInstance(object):

    def __init__(self, instance_info):
        self.student_id = instance_info['student_id']
        self.attendance_code = instance_info.get('attendance', {}).get('key')


class SyncAttendanceTask(JourneyCallTask):

    user_ids = None
    student_ids = None

    def default_last_sync(self, prefs):
        return datetime.now() - timedelta(days=prefs.days)

    def complete(self, request, result):
        prefs = IJourneyPreferences(self.context)
        prefs.last_sync = datetime.now()
        super(SyncAttendanceTask, self).complete(request, result)

    def execute(self, celery_task, *args, **kwargs):
        self.user_ids = {}
        self.student_ids = {}
        utility = getUtility(IJourneyAccessUtility)
        prefs = IJourneyPreferences(self.context)
        if prefs.last_sync is not None:
            dt = prefs.last_sync
        else:
            dt = self.default_last_sync(prefs)
        last_sync = dt.strftime('%Y-%m-%d')
        args = {
            'query[date.gte]': last_sync,
            'query[completed_at.null]': 'false',
        }
        data = urlencode(args)
        url = '%s/objects/register.json?%s' % (utility.account_url, data)
        request = self.request_factory(url)
        request.add_header('Authorization', utility.authorization_header)

        prefs = IJournalScoreSystemPreferences(ISchoolToolApplication(None))
        ss = prefs.student_scoresystem
        grades = {
            'present': 'P',
            'absent': 'AU',
        }
        try:
            response = urllib2.urlopen(request)
        except (urllib2.HTTPError,) as exc:
            raise
        else:
            decoder = JSONDecoder()
            register_infos_result = decoder.decode(response.read())
            for register_info in register_infos_result['objects']:
                register = Register(register_info)
                meeting_date = parse_date(register.date)
                meeting = makeSchoolAttendanceMeeting(meeting_date)
                requirement = StudentAMRequirement(meeting, ss)

                evaluator = None
                if register.user_id not in self.user_ids:
                    url = '%s/objects/user/%s.json' % (
                        utility.account_url, register.user_id)
                    request = self.request_factory(url)
                    request.add_header(
                        'Authorization', utility.authorization_header)
                    try:
                        response = urllib2.urlopen(request)
                    except (urllib2.HTTPError,) as exc:
                        raise
                    else:
                        decoder = JSONDecoder()
                        evaluator_result = decoder.decode(response.read())
                        person = self.context['persons'].get(
                            evaluator_result.get('schooltool_id'))
                        if person is not None:
                            evaluator = person.__name__
                        self.user_ids[register.user_id] = evaluator
                else:
                    evaluator = self.user_ids[register.user_id]

                args = {
                    'query[register_id]': register.id,
                }
                data = urlencode(args)
                url = '%s/objects/student_instance.json?%s' % (
                    utility.account_url, data)
                request = self.request_factory(url)
                request.add_header('Authorization', utility.authorization_header)
                try:
                    response = urllib2.urlopen(request)
                except (urllib2.HTTPError,) as exc:
                    raise
                else:
                    decoder = JSONDecoder()
                    student_instances_result = decoder.decode(response.read())
                    for instance_info in student_instances_result['objects']:
                        student_instance = StudentInstance(instance_info)
                        grade = grades.get(student_instance.attendance_code)
                        if grade is not None:
                            student = None
                            if student_instance.student_id not in self.student_ids:
                                url = '%s/objects/student/%s.json' % (
                                    utility.account_url, student_instance.student_id)
                                request = self.request_factory(url)
                                request.add_header('Authorization', utility.authorization_header)
                                try:
                                    response = urllib2.urlopen(request)
                                except (urllib2.HTTPError,) as exc:
                                    raise
                                else:
                                    decoder = JSONDecoder()
                                    student_result = decoder.decode(response.read())
                                    username = student_result['schooltool_id']
                                    student = self.context['persons'].get(username)
                                    self.student_ids[student_instance.student_id] = student
                            else:
                                student = self.student_ids[student_instance.student_id]
                            if student is not None:
                                score = IEvaluateRequirement(requirement).getEvaluation(
                                    student, requirement, default=UNSCORED)
                                if score is UNSCORED:
                                    IEvaluateRequirement(requirement).evaluate(
                                        student, requirement, grade,
                                        evaluator=evaluator)
